require 'rails_helper'
require "refile/file_double"

module RechnMgmWeb
    NoBrainer.sync_indexes

    RSpec.describe Api::V1::ApiDocumentsController, type: :controller do
        routes { RechnMgmWeb::Engine.routes }

        before(:each) {
            # Cleanup
            NoBrainer.purge!
            NoBrainer::Document::Index::Synchronizer.new([Nukleus::Document, DeviseAuthentication::User]).sync_indexes

            # Create some users
            ## Buchhaltung1 who is currently in his own organisation and has an admin role as well
            @buchhaltung = DeviseAuthentication::User.new(email: "buchhaltung@test.de", forename: "Buchhalter", lastname: "Eins", password: "123456", password_confirmation: "123456" )
            expect(@buchhaltung.save).to eq(true)
            @buchhaltung.member_ships.first.roles_mask = 7 # Buchhalter + Sachbearbeiter
            @organisation_1 = @buchhaltung.member_ships.where(is_primary: true).first.organisation
            @buchhaltung.set_organisation(@organisation_1)

            ## Buchhaltung2 who is currently in his own organisation
            @buchhaltung2 = DeviseAuthentication::User.new(email: "buchhaltung2@test.de", forename: "Buchhalter", lastname: "Zwei", password: "123456", password_confirmation: "123456" )
            expect(@buchhaltung2.save).to eq(true)
            @buchhaltung2.member_ships.first.roles_mask = 2
            @organisation_2 = @buchhaltung2.member_ships.where(is_primary: true).first.organisation
            @buchhaltung2.set_organisation(@organisation_2)

            ## Sachbearbeiter wo belongs to organisation 1 which is active right now.
            @sachbearbeiter = DeviseAuthentication::User.new(email: "sachbearbeiter@test.de", forename: "Sachbearbeiter", lastname: "Erster", password: "123456", password_confirmation: "123456" )
            expect(@sachbearbeiter.save).to eq(true)
            @sachbearbeiter_membership_to_organisation_1 = RechnMgmUserArea::MemberShip.new(user_id: @sachbearbeiter.id, organisation_id: @organisation_1.id, is_primary: false, roles_mask: 4)
            expect(@sachbearbeiter_membership_to_organisation_1.save).to be(true)
            @sachbearbeiter.current_membership_id = @sachbearbeiter_membership_to_organisation_1.id
            # set organisation_1 as current organisation
            @sachbearbeiter.set_organisation(@organisation_1)

            ### sachbearbeiter2 is member of two organisations; organisation 1 which is active right now.
            @sachbearbeiter2 = DeviseAuthentication::User.new(email: "sachbearbeiter2@test.de", forename: "Sachbearbeiter", lastname: "Zweiter", password: "123456", password_confirmation: "123456" )
            expect(@sachbearbeiter2.save).to eq(true)
            RechnMgmUserArea::MemberShip.create(user_id: @sachbearbeiter2.id, organisation_id: @organisation_1.id, is_primary: false, roles_mask: 4)
            RechnMgmUserArea::MemberShip.create(user_id: @sachbearbeiter2.id, organisation_id: @organisation_2.id, is_primary: false, roles_mask: 4)
            ### Work as member of organisation_1
            @sachbearbeiter2.set_organisation(@organisation_1)

            ## New User
            ### New user that belongs only to his own organisation. He is Admin, Buchhalter, and Sachbearbeiter
            @new_user = DeviseAuthentication::User.new(email: "new@test.de", forename: "New", lastname: "User", password: "123456", password_confirmation: "123456" )
            expect(@new_user.save).to eq(true)

            # Sachbearbeiter + Buchhalter..
            @sachbearbeiter_buchhalter =  DeviseAuthentication::User.new(email: "sachbearbeiter_buchhalter@test.de", forename: "Sachbearbeiter", lastname: "Buchhalter", password: "123456", password_confirmation: "123456" )
            expect(@sachbearbeiter_buchhalter.save).to eq(true)
            RechnMgmUserArea::MemberShip.create(user_id: @sachbearbeiter_buchhalter.id, organisation_id: @organisation_1.id, is_primary: false, roles_mask: 6)
            @sachbearbeiter_buchhalter.set_organisation(@organisation_1)

            # Simulate import of some new documents
            @new_document = create_document(@buchhaltung, "Test1")
            create_document(@buchhaltung, "Test2")
            create_document(@buchhaltung, "Test3")

            # Documenents are assigned to Sachbearbeiter
            @assigned_document = create_document(@sachbearbeiter, "Test4", nil, @buchhaltung)
            create_document(@sachbearbeiter, "Test5", Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], @buchhaltung)    # Unspecified (explicitly set)
            @accepted_document = create_document(@sachbearbeiter, "Test7", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung)       # Accepted
            create_document(@sachbearbeiter, "Test8", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung)         # Denied
            ## Some documents assigned to Sachbearbeiter from other organisation
            create_document(@sachbearbeiter, "Test9",  Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung2)         # Denied
            create_document(@sachbearbeiter, "Test10", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung2)         # Denied

            # Documenents are assigned to Sachbearbeiter2
            @assigned_document2 = create_document(@sachbearbeiter2, "Test4", nil, @buchhaltung)
            @doc_sachb_2_1 = create_document(@sachbearbeiter2, "Test5", Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], @buchhaltung)    # Unspecified (explicitly set)
            @doc_sachb_2_2 = create_document(@sachbearbeiter2, "Test7", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung)       # Accepted
            @assigned_document3 = create_document(@sachbearbeiter2, "Test8", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung)         # Denied
            @assigned_document4 = create_document(@sachbearbeiter2, "Test4", nil, @buchhaltung2)                                              # From Buchhaltung 2 (different Organisation)

            # Document is validated and back to Buchahltung
            create_document(@buchhaltung, "Test9", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung, true)       # Accepted
            create_document(@buchhaltung, "Test10", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung, true)         # Denied

            # Document is assigned to Buchhaltung (double role)
            create_document(@buchhaltung, "Test4", nil, @buchhaltung)
        }



        describe "Check Buchhaltung: " do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
            }

            it "finds new documents without validation state" do
                # New Documents
                get :index_unassigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(5) # 3 imported, 2 returned
                expect(body_data.select{|doc| doc["state"] == Nukleus::Document.document_states.index(:DOCUMENT_STATE_IMPORTED)}.count).to eq(3) # 3 are imported
                expect(body_data.select{|doc| doc["state"] == Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED)}.count).to eq(2) # 2 are returned
            end

            it "finds new documents with valdiation state Unspecified" do
                # New Documents
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(3)
                expect(body_data.map{|d| d["organisationId"]}).not_to include( be_nil )
            end

            it "finds documents assigned to sachbearbeiters (in progress)" do
                get :index_assigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(8)
            end

            it "finds documents assigned to sachbearbeiter that are not yet validated (in progress, unvalidated)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(4)
            end

            it "finds documents assigned to other person but are validated (in progess, accepted, not returned)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "finds documents assigned to other person but are validated (in progess, denied, not returned)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "finds documents given back to buchhaltung and validated (finished, denied)" do
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED] } , format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
            end

            it "finds documents given back to buchhaltung but are validated (finished and accepted)" do
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
            end

            it "finds all documents he has to work for (assigned)" do
                get :index, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
            end

            it "finds all validated documents" do
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "receives a status overview" do
                get :status_overview, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)

                expect(body_data.keys.include? "new").to be(true)
                expect(body_data["new"].count).to eq(4)             # 3 Imported, 1 Assigned
                expect(body_data["new"].first.class).to be(String)

                expect(body_data.keys.include? "assigned").to be(true)
                expect(body_data["assigned"].count).to eq(8)
                expect(body_data["assigned"].first.class).to be(String)

                expect(body_data.keys.include? "finished").to be(true)
                expect(body_data["finished"].count).to eq(2)        # 2 Returned
                expect(body_data["finished"].first.class).to be(String)
            end

            it "receives status of other users for organisation_1" do
                @buchhaltung.set_organisation(@organisation_1)

                get :status_user, :params => {user_ids: "#{@sachbearbeiter.id},#{@sachbearbeiter2.id}"}, format: :json

                expect(response).to have_http_status(:ok)
                body_data = JSON.parse(response.body)

                expect(body_data.class).to be(Hash)
                expect(body_data.keys.include? @sachbearbeiter.id).to be(true)
                expect(body_data["#{@sachbearbeiter.id}"].count).to eq(4)
                expect(body_data.keys.include? @sachbearbeiter2.id).to be(true)
                expect(body_data["#{@sachbearbeiter2.id}"].count).to eq(4)
            end

            it "set validation state" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment")
            end

            it "set validation state of freshly imported document" do
                expect(@new_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(@new_document.organisation_id).not_to be(nil)

                put :set_validation_state, :params => {id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@new_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment")
                expect(test_doc.organisation_id).not_to be(nil)
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_IMPORTED))
            end

            it "assigns document to user" do
                expect(@new_document.user_id).to eq(@buchhaltung.id)
                expect(@new_document.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_IMPORTED))

                put :assign_to_user, :params => { id: @new_document, user_id: @sachbearbeiter2, comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@new_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
                # Document must now be assigned to organisation_1
                expect(test_doc.organisation_id).to eq(@organisation_1.id)
            end

            it "validates a document" do
                expect(@new_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@new_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
            end

            it "validates a document and assigns to user in one step" do
                expect(@new_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state_and_assign, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment", user_id_of_destination: @sachbearbeiter2.id }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@new_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("test-comment")
                expect(test_doc.validation).to eq(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
            end

            it "validates whether buchhalter inbox call behaves correctly" do
                # Inbox of buchhalter adds imported documents and assigned socuments

                # Get number of assigned documents via index / all_by_user
                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).not_to eq(0)
                assigned_documents = body_data.count

                # Get number of imported documents via index_unassigned / unassigned_by_validation
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).not_to eq(0)
                imported_documents = body_data.count

                # Check whether the number from buchhalter_inbox is correct.
                get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to have_http_status(:ok)
                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(assigned_documents + imported_documents)
            end
        end

        describe "Check Sachbearbeiter 1: " do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
            }

            it "sees only documents assigned to myself" do
                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(4)
            end

            it "set validation state" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
            end

            xit "cannot set validation state if not assigned to him" do
                expect(@assigned_document2.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document2, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).not_to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document2.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.history.count).to eq(1)
            end

            it "does not receives a status overview" do
                expect{
                    get :status_overview, { format: :json }
                }.to raise_error("buchhalter only")
            end

            it "assigns document to user" do
                expect(@assigned_document.user_id).to eq(@sachbearbeiter.id)
                expect(@assigned_document.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))

                put :assign_to_user, :params => { id: @assigned_document, user_id: @sachbearbeiter2, comment: "My Comment 3" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment 3")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
            end

            it "validates a document and assigns to user in one step" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state_and_assign, :params => { id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment", user_id_of_destination: @sachbearbeiter2.id }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("test-comment")
                expect(test_doc.validation).to eq(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
            end

            it "validates a document and returns to buchhalter in one step" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state_and_assign, :params => { id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment2", back_to_origin: "true" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@buchhaltung.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("test-comment2")
                expect(test_doc.validation).to eq(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED))

                # Document must be in "ready to export" only of buchhalter
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                is_document_in_response_body?(@assigned_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                is_document_in_response_body?(@assigned_document, true) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                is_document_in_response_body?(@assigned_document, false) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end

            it "only returns to buchhalter a set to unvalidated in one step" do
                expect(@accepted_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED])

                put :set_validation_state_and_assign, :params => { id: @accepted_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], comment: "test-comment2", back_to_origin: "true" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@accepted_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@buchhaltung.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("test-comment2")
                expect(test_doc.validation).to eq(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED))

                # Document must be in "ready to export" only of buchhalter
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                # Doc is not assigned anymore to other sachbearbeiter
                is_document_in_response_body?(@accepted_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Not assigned because validaton is unspecified.
                is_document_in_response_body?(@accepted_document, false) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@accepted_document, true) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end

            it "only returns to buchhalter (with set to unvalidated) in one step" do
                expect(@accepted_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED])

                put :set_back_to_origin_and_set_state, :params => { id: @accepted_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], comment: "test-comment2" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@accepted_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@buchhaltung.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("test-comment2")
                expect(test_doc.validation).to eq(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED))

                # Document must be in "ready to export" only of buchhalter
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                # Doc is not assigned anymore to other sachbearbeiter
                is_document_in_response_body?(@accepted_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Not assigned because validaton is unspecified.
                is_document_in_response_body?(@accepted_document, false) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@accepted_document, true) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end

            xit "can read document assigned to him" do
                get :show, :params => {id: @assigned_document.id }, format: :json
                expect(response).to have_http_status(:ok)
            end

            xit "cannot read document not assidned to him" do
                get :show, :params => {id: @assigned_document2.id }, format: :json
                expect(response).not_to have_http_status(:ok)
            end

            xit "cannot read preview document not assidned to him" do
                get :preview, :params => { id: @assigned_document2.id, width: "200", height:"400" },  format: :json
                expect(response).not_to have_http_status(:ok)
            end

        end

        describe "Check Sachbearbeiter 2: " do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter2)
            }

            it "should respect organisation with index " do
                # sachbearbeiter 2 sees his assigned documents only if he is in the context of the right organisation

                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(4)

                # Check that the correct documents are assigned, assigned_document4 should not be found
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document2.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @doc_sachb_2_1.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @doc_sachb_2_2.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document3.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document4.id}.count).to eq(0)

                # Now switch to organisation 2 and check that there is only 1 documents visible and is assigned_document4
                @sachbearbeiter2.set_organisation(@organisation_2)
                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document4.id}.count).to eq(1)
            end

            it "should respect organisation with index_assigned " do
                # sachbearbeiter 2 sees his assigned documents only if he is in the context of the right organisation

                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(4)

                # Check that the correct documents are assigned, assigned_document4 should not be found
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document2.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @doc_sachb_2_1.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @doc_sachb_2_2.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document3.id}.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document4.id}.count).to eq(0)

                # Now switch to organisation 2 and  check that there is only 1 documents visible and is assigned_document4
                @sachbearbeiter2.set_organisation(@organisation_2)
                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(1)
                expect(body_data.map { |d| d["id"] }.select{|id| id == @assigned_document4.id}.count).to eq(1)
            end
        end

        describe "Check Sachbearbeiter and Buchhaltung" do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter_buchhalter)
            }

            it "does not see validated documents as new" do
                get :status_overview, { format: :json }
                expect(response).to have_http_status(:ok)
                resp_data = JSON.parse(response.body)
                expect(find_duplicates(resp_data["new"] + resp_data["finished"]).nil?).to be(true)
            end
        end

        describe "Check Document assignment workflows" do
            render_views
            it "return document back from sachbearbeiter to buchhaltung" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document will not be found
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to eq(false)

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @accepted_document.id, comment: "My Comment2" }, format: :json
                expect(response).to have_http_status(:ok)

                # Check that this document will now be found
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to be(true)

                # Check that comment was added.
                test_doc = Nukleus::Document.find(@accepted_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment2")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED))
            end

            # index_validated == validatedDocuments
            it "return document back to buchhaltung for validatedDocuments" do
                # Verwaltung is doing this call. Mock current user..
                @buchhaltung.set_organisation(@organisation_1)
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document will not be found
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to eq(false)

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @accepted_document.id }, format: :json
                expect(response).to have_http_status(:ok)

                # Check that this document will now be found
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(3)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to be(true)
            end

            it "return document back to buchhaltung for not validatedDocuments" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                ## Check that this document will not be found by buchhaltung (precondition)
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to eq(false)

                ## Now do the back_to_orign operation
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @assigned_document.id }, format: :json
                expect(response).to have_http_status(:ok)

                # Switch unser to Buchhaltung
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                ## Check that this document cannot be found as validated document
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to be(false)

                ## Check that the document can be found as "new"
                get :index_unassigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to be(true)
            end

            it "return documents back to buchhaltung if a sachbearbeiter is deleted" do
                [@sachbearbeiter, @sachbearbeiter2].each do |user|
                    allow(controller).to receive(:current_user).and_return(user)
                    get :index_as_ids, { format: :json }
                    expect(response).to have_http_status(:ok)

                    assigned_documents = JSON.parse(response.body)
                    expect(assigned_documents.count).not_to eq(0)

                    #Now delete the user. All documents should now be assigned back to its origin.
                    user.destroy

                    doc_count = 0
                    assigned_documents.each do |document_id|
                        doc = Nukleus::Document.find(document_id)
                        expect(doc.user_id).to eq(doc.origin_user_id)
                        doc_count += 1
                    end

                    expect(Nukleus::Document.where(user_id: user.id).count).to be(0)
                    expect(doc_count).not_to eq(0)

                    # No memberships should be left
                    expect(RechnMgmUserArea::MemberShip.where(user_id: user.id).count).to eq(0)
                end
            end

            it "delete documents if a buchhalter is deleted" do
                [@buchhaltung].each do |user|
                    allow(controller).to receive(:current_user).and_return(user)

                    #Now delete the user. All documents should now be assigned back to its origin.
                    user.destroy

                    doc_count = 0
                    Nukleus::Document.each do |doc|
                        expect(doc.origin_user_id).not_to eq(user.id)
                        expect(doc.user_id).not_to eq(user.id)
                        doc_count += 1
                    end

                    expect(doc_count).not_to eq(0)
                end
            end

            it "return documents to buchhaltung if membership is deleted" do
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                # Sachbearbeiter 1 is assigned to buchhalter's organsiation and owns a document.
                expect(@sachbearbeiter.id).to eq(@assigned_document.user_id)

                # delete membership with running events
                expect(@sachbearbeiter_membership_to_organisation_1.destroy).to eq(true)

                # Document must now be back to buchhaltung
                check_document = Nukleus::Document.find(@assigned_document.id)
                expect(@buchhaltung.id).to eq(check_document.user_id)
            end
        end

        describe "Check buchhalter workflows as executed by frontend calls: " do
            render_views

            before(:each){
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
            }

            it "accept freshly imported" do
                # This is called by the frontend if buchhalter accepts or denies a document.
                put :set_validation_state_and_assign, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment2", back_to_origin:true }, format: :json
                expect(response).to have_http_status(:ok)

                # Document must be in "ready to export" only of buchhalter
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@new_document, false) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Doc is not assigned to other sachbearbeiter
                is_document_in_response_body?(@new_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # It is validated .
                is_document_in_response_body?(@new_document, true) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end

            it "accept freshly imported and assign to sacbearbeiter, sachbearbeiter returns and accepts" do
                # This is called by the frontend if buchhalter accepts or denies a document.
                put :set_validation_state_and_assign, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment2", user_id_of_destination: @sachbearbeiter.id }, format: :json
                expect(response).to have_http_status(:ok)

                # Document must be in "ready to export" only of buchhalter
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@new_document, false) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Doc is not assigned to other sachbearbeiter
                is_document_in_response_body?(@new_document, true) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # It is validated .
                is_document_in_response_body?(@new_document, false) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :set_validation_state_and_assign, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment2", back_to_origin: "true" }, format: :json
                expect(response).to have_http_status(:ok)

                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                # Document must be in "ready to export" only of buchhalter
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@new_document, false) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Doc is not assigned to other sachbearbeiter
                is_document_in_response_body?(@new_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # It is validated .
                is_document_in_response_body?(@new_document, true) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end

            it "accept freshly imported and assign to sacbearbeiter, sachbearbeiter returns without accepting or denying" do
                put :set_validation_state_and_assign, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "test-comment2", user_id_of_destination: @sachbearbeiter.id }, format: :json
                expect(response).to have_http_status(:ok)

                # Document must be in "ready to export" only of buchhalter
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@new_document, false) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Doc is not assigned to other sachbearbeiter
                is_document_in_response_body?(@new_document, true) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # It is validated .
                is_document_in_response_body?(@new_document, false) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :set_back_to_origin_and_set_state, :params => { id: @new_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], comment: "test-comment2" }, format: :json
                expect(response).to have_http_status(:ok)

                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                # Document must be in "ready to export" only of buchhalter
                # Should be back to inbox due to unspecified
                is_document_in_response_body?(@new_document, true) do
                    get :buchhalter_inbox, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # Doc is not assigned to other sachbearbeiter
                is_document_in_response_body?(@new_document, false) do
                    get :index_assigned, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
                # It is validated .
                is_document_in_response_body?(@new_document, false) do
                    get :index_validated, { format: :json }
                    expect(response).to have_http_status(:ok)
                    response.body
                end
            end
        end

        describe "Document basic stuff" do
            render_views
            it "get preview" do
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                document = Nukleus::Document.first
                get :preview, :params =>  {id: document.id, width: 600, height: 800}
                expect(response).to have_http_status(:ok)
            end

            it "has a history" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)

                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 1")).to be(true)
                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 2")).to be(true)
                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 3")).to be(true)

                get :show, :params => { id: @assigned_document.id },  format: :json
                expect(response).to render_template("show")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data["history"].count).to eq(3)
            end

            it "add a comment" do
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)

                initial_history_count = @assigned_document.history.count

                put :add_comment, :params => { id: @assigned_document, comment: "Test" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc.history.count).to eq(initial_history_count + 1)
                expect(test_doc.history.last["comment"]).to eq("Test")
            end

            # TODO: Add testing for action cable
            # it "created will be signalled with action cable" do
            #     expect { put :add_comment, :params => { id: @assigned_document, comment: "Test" }, format: :json }.to have_broadcasted_to("document_channel") #.with(text: 'Cool!')
            # end
        end

        # Buchhalter 3 is lonely and has no connection to anyone.
        describe "Documents for lonely new user" do
            render_views

            before(:each){
                allow(controller).to receive(:current_user).and_return(@new_user)
            }

            it "index does not see any document" do
                get :index, { format: :json }

                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)
                body_data = JSON.parse(response.body)

                expect(body_data.count).to eq(0)
            end

            it "index_unassigned does not see any document" do
                get :index_unassigned, { format: :json }

                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)
                body_data = JSON.parse(response.body)

                expect(body_data.count).to eq(0)
            end

        end


        private

        def create_document(user, file_name, validation_state = nil, origin_user = nil, is_returned = false)
            backend = Refile.backends["cache"]

            test_file = File.open("./spec/data/test.pdf")
            file = backend.upload(test_file)

            params = {
                document_file:              file,
                document_id:                SecureRandom.uuid,
                document_file_filename:     file_name.gsub( /[^a-zA-Z0-9_\.]/, '_'),
                # document_file_size:         file.size,
                # document_file_content_type: type,
                user_id:                    user.id,
                meta_data:                  { importer_type: :TEST },
                organisation_id: 			user.current_organisation.id
            }

            if origin_user.present?
                # Document is assigned. Thus we simulate an assignment
                source_organisation_id   = origin_user.my_organisation.id
                params[:origin_user_id]  = origin_user.id
                params[:organisation_id] = source_organisation_id
                if is_returned
                    params[:state] = Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED)
                else
                    params[:state] = Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED)
                end
            end

            if validation_state.present?
                params[:validation] = validation_state
            end

            new_doc = Nukleus::Document.new(params)
            expect(new_doc.save).to eq(true)

            new_doc
        end

        def is_document_in_response_body?(document, expected_found = true)
            body_data = JSON.parse(yield)
            expect(body_data.map { |d| d["id"] }.select{|id| id == document.id}.count).to eq(expected_found ? 1 : 0)
        end

        def find_duplicates(arr)
            s = Set.new
            arr.find { |e| !s.add?(e) }
        end
    end

end
